package aplprolog.prolog;

import aplprolog.datastructures.PrimIntArrayList;
import aplprolog.prolog.builtins.*;
/**
 * Built-in Prolog functions. Basically anything that can be (or must be) done Java. This is generally faster
 * then building them in a library.
 * 
 * @author Bas Testerink, Utrecht University, The Netherlands
 *
 */
public class BuiltIn {
	// interceptors (intercept resolution to execute a build in method)
	public static final int IC_UNIFY=0,IC_NOTUNIFY=1,IC_VAR=2,IC_ATOM=3,IC_INTEGER=4,IC_FLOAT=5,IC_ATOMIC=6,IC_COMPOUND=7,IC_NONVAR=8,IC_NUMBER=9,IC_TERMEQLT=10,IC_TERMEQ=11,IC_TERMNEQ=12,IC_TERMLT=13,IC_TERMGT=14,IC_TERMGTEQ=15,IC_FUNCTOR=16,IC_ARG=17,IC_DECOMPOSE=18,IC_COPY_TERM=19,IC_IS=20,IC_ARITHEQ=21,IC_ARITHNEQ=22,IC_LT=23,IC_EQLT=24,IC_GT=25,IC_GTEQ=26,IC_CLAUSE=27,IC_CURPRED=28,IC_ASSERTA=29,IC_ASSERTZ=30,IC_ASSERT=31,IC_UNIQUEASSERTZ=32,IC_UNIQUEASSERTZ2=33,IC_ATOMIC_UPDATE=34,IC_RETRACT=35,IC_RETRACTALL=36,IC_ABOLISH=37,IC_FINDALL=38,IC_BAGOF=39,IC_SETOF=40,IC_ONCE=41,IC_REPEAT=42,IC_ATOM_LENGTH=43,IC_ATOM_CONCAT=44,IC_SUB_ATOM=45,IC_ATOM_CHARS=46,IC_ATOM_CODES=47,IC_CHAR_CODE=48,IC_NUMBER_CHARS=49,IC_NUMBER_CODES=50,IC_RANDOM=51,IC_TRUE=52,IC_INT_TO_ATOM=53,IC_FAIL=54,IC_BUILT_IN_COLLECTOR_DO_NOT_USE=55,IC_EXTERNAL=56,IC_EXTERNAL1=57,IC_EXTERNAL2=58,IC_RESETSTABLE=59,IC_ISSTABLE=60,IC_RULEIMPLY=61;
	String[] intercepts = new String[]{"=","\\=","var","atom","integer","float","atomic","compound","nonvar","number","@=<","==","\\==","@<","@>","@>=","functor","arg","=..","copy_term"," is ","=:=","=\\=","<","=<",">",">=","clause","current_predicate","asserta","assertz","assert","uniqueassertz","uniqueassertz2","atomic_update","retract","retractall","abolish","findall","bagof","setof","once","repeat","atom_length","atom_concat","sub_atom","atom_chars","atom_codes","char_code","number_chars","number_codes","random","true","int_to_atom","fail","IC_BUILT_IN_COLLECTOR_DO_NOT_USE","@external3","@external","@external2","built_in_reset_stable_flag","built_in_is_stable",":-"};
	public static final int[] arity = new int[]{2,2,1,1,1,1,1,1,1,1,2,2,2,2,2,2,3,3,2,2,2,2,2,2,2,2,2,2,1,1,1,1,1,1,3,1,1,1,3,3,3,1,0,2,3,5,2,2,2,2,2,1,0,2,0,2,4,3,2,0,0,2};
	// functions
	public static final int ADD=0,SUBTRACT=1,MULTIPLY=2,DIVIDE=3,INTDIVIDE=4,REMAINDER=5,MODULO=6,ABS=7,SIGN=8,INTEGERPART=9,FRACPART=10,FLOAT=11,FLOOR=12,CEIL=13,CEILING=14,TRUNCATE=15,ROUND=16,STARPOW=17,POW=18,SIN=19,COS=20,ATAN=21,EXP=22,LOG=23,LOG10=24,SQRT=25,SHIFTRIGHT=26,SHIFTLEFT=27,AND=28,OR=29,COMPLEMENT=30,MAX=31,MIN=32,INTEGER=33,XOR=34,TAN=35,ASIN=36,ACOS=37,PI=38,E=39,ENQUEUED=40,IMMEDIATE=41;
	String[] functions = new String[]{"+","-","*","/"," div "," rem "," mod ","abs","sign","float_integer_part","float_fractional_part","float","floor","ceil","ceiling","truncate","round","**","^","sin","cos","atan","exp","log","log10","sqrt",">>","<<","/\\","\\/","\\","max","min","integer"," xor ","tan","asin","acos","pi","e","enqueued","immediate"};

	public PrimIntArrayList externalCallQueue = new PrimIntArrayList(false);
	public int range = 0;
	public Prolog p;
	public Arithmetics arithmetics;
	public AtomBuilding atombuilding;
	public AnswerSets answersets;
	public RulebaseModifications rulemods;
	public ExternalActions external;
	/*
	 * Built-in types:
	 * - intercept built-ins ("is","var","num" etc)
	 * - function built-ins ("+","abs" etc)
	 * - rule built-ins (";", "member", etc)
	 */
	public BuiltIn(Prolog p){
		this.p = p;
		arithmetics = new Arithmetics(this);
		atombuilding = new AtomBuilding(this);
		answersets = new AnswerSets(this); 
		rulemods = new RulebaseModifications(this);
		external = new ExternalActions(this);
		range = intercepts.length;
		for(String s : intercepts)p.strStorage.add(s);
		for(String s : functions)p.strStorage.add(s);
		if(p.trace)System.out.println();
	}
	/*
	 * Big switch to map call id to correct java call. hasBeenCalled can be used to track whether 
	 * the built-in predicate is backtrackable (generally not). Various functions are distributed
	 * in the builtin package to keep the code somewhat readable.
	 */
	/**
	 * Execute a built-in predicate.
	 * @param call A call that is a built-in predicate.
	 * @return Whether execution was successful. 
	 */
	public boolean execute(int[] call, boolean invar){
		int id = call[1];
		arithmetics.arithcursor = 3;
		switch(id){
		case IC_UNIFY:		return (hasBeenCalled(call)?false:(unifycall(call,invar)?true:false));
		case IC_NOTUNIFY:	return (hasBeenCalled(call)?false:(unifycall(call,invar)?false:true));
		case IC_VAR:		return (hasBeenCalled(call)?false:(call[3]==IntHarvester.VARIABLE?p.substitution.value(call[4])==null:false));
		case IC_ATOM:		return (hasBeenCalled(call)?false:atombuilding.isAtom(call,3,false));
		case IC_INTEGER:	double d = arithmetics.evaluate(call);if(Double.isNaN(d)) return false;return arithmetics.isInt(d);
		case IC_FLOAT:		return (hasBeenCalled(call)?false:!Double.isNaN(arithmetics.evaluate(call)));
		case IC_ATOMIC:		return (hasBeenCalled(call)?false:(atombuilding.isAtom(call,3,false)||!Double.isNaN(arithmetics.evaluate(call))));
		case IC_COMPOUND:	return (hasBeenCalled(call)?false:(!(atombuilding.isAtom(call,3,false)||!Double.isNaN(arithmetics.evaluate(call))) && !(call[3]==IntHarvester.VARIABLE?p.substitution.value(call[4])==null:false)));
		case IC_NONVAR:		return (hasBeenCalled(call)?false:(call[3]==IntHarvester.VARIABLE?p.substitution.value(call[4])!=null:false));
		case IC_NUMBER:		return (hasBeenCalled(call)?false:!Double.isNaN(arithmetics.evaluate(call)));
		case IC_TERMEQLT:	return (hasBeenCalled(call)?false:compareTerm(call, id));
		case IC_TERMEQ:		return (hasBeenCalled(call)?false:compareTerm(call, id));
		case IC_TERMNEQ:	return (hasBeenCalled(call)?false:compareTerm(call, id));
		case IC_TERMLT:		return (hasBeenCalled(call)?false:compareTerm(call, id));
		case IC_TERMGT:		return (hasBeenCalled(call)?false:compareTerm(call, id));
		case IC_TERMGTEQ:	return (hasBeenCalled(call)?false:compareTerm(call, id));
		case IC_FUNCTOR:	return false;
		case IC_ARG:		return false;
		case IC_DECOMPOSE:	return false;
		case IC_COPY_TERM:	return false;
		case IC_IS:		return (hasBeenCalled(call)?false:arithmetics.arithmetic(call));
		case IC_ARITHEQ:	return (hasBeenCalled(call)?false:arithmetics.arithmeticEQ(call));
		case IC_ARITHNEQ:	return (hasBeenCalled(call)?false:!arithmetics.arithmeticEQ(call));
		case IC_LT:		return (hasBeenCalled(call)?false:compareNum(call, id));
		case IC_EQLT:		return (hasBeenCalled(call)?false:compareNum(call, id));
		case IC_GT:		return (hasBeenCalled(call)?false:compareNum(call, id));
		case IC_GTEQ:		return (hasBeenCalled(call)?false:compareNum(call, id));
		case IC_CLAUSE:		return false;
		case IC_CURPRED:	return false;
		case IC_ASSERTA:	return hasBeenCalled(call)?false:rulemods.queryAssert(call,true,false,false);
		case IC_ASSERTZ:	return hasBeenCalled(call)?false:rulemods.queryAssert(call,false,false,false);
		case IC_ASSERT:		return hasBeenCalled(call)?false:rulemods.queryAssert(call,false,false,false);
		case IC_UNIQUEASSERTZ:	return hasBeenCalled(call)?false:rulemods.queryAssert(call,false,true,false);
		case IC_UNIQUEASSERTZ2:	return hasBeenCalled(call)?false:rulemods.queryAssert(call,false,true,true);
		case IC_ATOMIC_UPDATE:	return hasBeenCalled(call)?false:rulemods.atomicUpdate(call);
		case IC_RETRACT:	return hasBeenCalled(call)?false:rulemods.queryRetract(call,false);
		case IC_RETRACTALL:	return hasBeenCalled(call)?false:rulemods.queryRetract(call,true);
		case IC_ABOLISH:	return false;
		case IC_FINDALL:	return answersets.collect(IC_FINDALL,call);
		case IC_BAGOF:		return answersets.collect(IC_BAGOF,call);
		case IC_SETOF:		return answersets.collect(IC_SETOF,call);
		case IC_ONCE:		return false;
		case IC_REPEAT:		return false;
		case IC_ATOM_LENGTH:	return false;
		case IC_ATOM_CONCAT:	return (hasBeenCalled(call)?false:atombuilding.concatAtom(call));
		case IC_SUB_ATOM:	return false;
		case IC_ATOM_CHARS:	return false;
		case IC_ATOM_CODES:	return false;
		case IC_CHAR_CODE:	return false;
		case IC_NUMBER_CHARS:	return false;
		case IC_NUMBER_CODES:	return false;
		case IC_RANDOM:		return (hasBeenCalled(call)?false:randomNum(call));
		case IC_TRUE:		return !hasBeenCalled(call);
		case IC_INT_TO_ATOM:	return (hasBeenCalled(call)?false:atombuilding.intToAtom(call, 3));
		case IC_FAIL:		return false;
		case IC_BUILT_IN_COLLECTOR_DO_NOT_USE:	answersets.expandCollection(call); return false;
		case IC_EXTERNAL:	return (hasBeenCalled(call)?false:externalCall(call));
		case IC_EXTERNAL1:	return (hasBeenCalled(call)?false:external.call(call,true));
		case IC_EXTERNAL2:	return (hasBeenCalled(call)?false:external.call(call,false));
		case IC_RESETSTABLE: p.rulesModified = false; return true;
		case IC_ISSTABLE:	return !p.rulesModified;
		case IC_RULEIMPLY:	return false;
		}
		if(p.trace)System.out.println(": "+p.strStorage.getString(id)+" not (yet) implemented");
		return false;
	}
	
	/**
	 * Check, on the basis of the resolution history, whether a call has been called before.
	 * @param call The call to be checked.
	 * @return Whether the call has been called before.
	 */
	public boolean hasBeenCalled(int[] call){
		if(p.history.intdata[p.history.pointer-2]>0){
			if(p.trace)System.out.println(": cannot redo "+p.harvester.arToString(call, false));
			return true;// cannot redo arithmetics
		}
		return false;
	}
	
	/**
	 * Unify the two arguments.
	 * @param call Call to the "=" operator.
	 * @param invar Whether the call was inside a variable.
	 * @return The result. 
	 */
	public boolean unifycall(int[] call, boolean invar){ 
		if(p.harvester.unifyTerm(call, call, 3, p.harvester.scanElement(call, 3, false, invar), invar, invar))
			return true;
		else {
			p.backtrack(false);
			return false;
		}
	}
	
	/**
	 * Generate a random number. If there is an bound variable in the call, then it gets instantiated with the random value.
	 * @param call Of the form random(X), where X is a unbound var (otherwise the call fails).
	 * @return Whether the call is successful. 
	 */
	public boolean randomNum(int[] call){
		if(call[3]==IntHarvester.VARIABLE){
			if(p.substitution.value(call[4])==null){
				double num = Math.random();
				int a = (int)((Double.doubleToLongBits(num)>>>32));
				int b = (int)((Double.doubleToLongBits(num)<<32)>>>32);
				p.substitution.bind(call[4], new int[]{IntHarvester.NUMBER,a,b});
				return true;
			}
		}
		return false;
	}
	/**
	 * Various term comparisons (such as "@>")
	 * @param call Comparison call.
	 * @param comparison Comparison type.
	 * @return Comparison result.
	 */
	public boolean compareTerm(int[] call, int comparison){
		int[] arg1 = call;
		int arg1c = 3;
		if(call[3]==p.harvester.VARIABLE){
			arg1 = p.substitution.value(call[4]);
			if(arg1 == null)return false;
			arg1c = 0;
		} else if(call[3] != p.harvester.PREDICATE) return false;
		
		int[] arg2 = call;
		int arg2c = p.harvester.scanElement(call, 3, false, false);
		if(call[arg2c]==p.harvester.VARIABLE){
			arg2 = p.substitution.value(call[arg2c+1]);
			if(arg2 == null)return false;
			arg2c = 0;
		} else if(call[arg2c] != p.harvester.PREDICATE) return false;
		 
		p.harvester.showVarIDs = false;
		String e1 = p.harvester.elemToString(arg1, arg1c, false);
		String e2 = p.harvester.elemToString(arg2, arg2c, false); 
		p.harvester.showVarIDs = true;
		switch(comparison){
		case IC_TERMEQ: return e1.equals(e2);
		case IC_TERMGT: return e1.compareTo(e2)>0;
		case IC_TERMGTEQ: return e1.compareTo(e2)>=0;
		case IC_TERMLT: return e1.compareTo(e2)<0;
		case IC_TERMEQLT: return e1.compareTo(e2)<=0;
		case IC_TERMNEQ: return !e1.equals(e2);
		}
		return false;
	}

	/**
	 * Various number comparisons (such as ">")
	 * @param call Comparison call.
	 * @param comparison Comparison type.
	 * @return Comparison result.
	 */
	public boolean compareNum(int[] call, int comparison){
		arithmetics.arithcursor = 3;
		double num1 = arithmetics.evaluate(call);
		double num2 = arithmetics.evaluate(call);
		if(Double.isNaN(num1)||Double.isNaN(num2)) return false; 
		
		switch(comparison){
		case IC_GT: if(p.trace)System.out.println(": "+num1+">"+num2+"?"); return num1 > num2;
		case IC_EQLT: if(p.trace)System.out.println(": "+num1+"=<"+num2+"?"); return num1 <= num2; 
		case IC_GTEQ: if(p.trace)System.out.println(": "+num1+">="+num2+"?"); return num1 >= num2; 
		case IC_LT: if(p.trace)System.out.println(": "+num1+"<"+num2+"?"); return num1 < num2;
		}
		return false;
	}
	
	/**
	 * Perform an external call. (made for standard 2OPL)
	 * @param call The call to the function.
	 */
	public boolean externalCall(int[] call){
		int[] dummycopy = rulemods.deepcopy(call, 0, call.length);
		String environment = p.strStorage.getString(dummycopy[4]);
		String actionStr = p.harvester.elemToString(dummycopy, 6, false);
		
		int action = p.harvester.scanElement(call, 3, false, false);
		int exectimeCursor = p.harvester.scanElement(call, action, false, false)+1;
		int exectime = call[exectimeCursor];
		if(exectime == range+ENQUEUED){ // enqueue until after sanction phase
			if(true){
				System.out.println(": Enqueue external action "+actionStr+" on environment "+environment+". (not yet implemented)");
			}
			rulemods.addZeroSplit = true;
			int[] copy = rulemods.deepcopy(call, 0, call.length);
			rulemods.addZeroSplit = false; 
			exectimeCursor = p.harvester.scanElement(copy, p.harvester.scanElement(call, 3, false, false), false, false)+1;
			copy[exectimeCursor] = range+IMMEDIATE; // next time do it immediately 
			externalCallQueue.addAr(copy); // enqueue the action 
		} else {
			if(p.trace){
				System.out.println(": Perform external action "+actionStr+" on environment "+environment+". (not yet implemented, so returns true)");
			}
		}
		return true;
	}
}
