package aplprolog.prolog; 

import aplprolog.datastructures.PrimIntArrayList; 
import aplprolog.datastructures.PrimIntKeyHashMap;
/**
 * Support class for the resolution engine. Does all the unification of integers and can extract data from integer arrays.
 * 
 * @author Bas Testerink, Utrecht University, The Netherlands
 *
 */
public class IntHarvester {
	public final static int CUT=-1,PREDICATE=-2,VARIABLE=-3,NUMBER=-4,ATOM=-5,LIST=-6,ARITHMETIC=-7,OPERATOR=-8,PARENTHESIZED=-9,UNIFICATION=-10;
	public final static int TAIL=-1;
	Prolog p;
	PrimIntArrayList array1 = new PrimIntArrayList(true);
	PrimIntArrayList body = new PrimIntArrayList(false);

	// deep copy mode settings (for assertion/retraction/findall/setof/bagof)
	public boolean deepcopymode=false; 
	public int deepcopycursor = 0;

	public IntHarvester(Prolog p){
		this.p = p;
	}

	/**
	 * Split a rule without split data overhead integers. 
	 * @param rule Rule to be split.
	 * @param cursor Start index.
	 * @return List of arrays containing the subgoals.
	 */
	public PrimIntArrayList getBody3(int[] rule, int cursor){
		body.clear(); 
		int a = cursor;
		int b = 0;
		while(b!=rule.length){ 
			b = scanElement(rule, a, true, false);
			body.addAr(p.extractArray(rule, a, b));
			a=b;
		}
		return body;
	}
	/**
	 * Use the split data to turn the subgoals of a rule/query into separate calls.
	 * @param rule Rule to be split.
	 * @return List of integer arrays.
	 */
	public PrimIntArrayList getBody2(int[] rule){
		body.clear(); 
		int splits = rule[rule.length-1];
		if(splits==1) return body; // no body
		int size = 0;
		int stopIndex = 0; 

		int c = rule[rule.length-2]; // move cursor behind the head
		int base = rule.length-2; // base start
		for(int s = 1; s < splits; s++){ // for each body part, except the head (hence s = 1)
			stopIndex = rule[base-s]; // index of the first integer after the bodypart
			size = stopIndex-c; // the array size of the bodypart 
			int[] newGoal = new int[size];
			for(int i = c; i < stopIndex; i++){
				newGoal[i-c] = rule[i]; // copy numbers
			}
			//System.arraycopy(rule, c, newGoal, 0, size);
			body.addAr(newGoal);
			c = stopIndex;
		}
		return body;
	}
	/**
	 * Replace the local vars in the rule with global variables.
	 * @param rule Rule to be modified.
	 */
	public void buildLocalVars(int[] rule){ 
		int c = rule.length-rule[rule.length-1]-2; // init as var cursor
		int nrVars = rule[c]; // the amount of local variables
		int varID = 0; // current id (is a counter)
		int occurences = 0; // occurence counter
		if(nrVars > 0){// replace the vars
			for(int i = 0; i < nrVars; i++){ // for each variable
				c--; // move cursor to the variable occurence counter
				occurences = rule[c]; // amount of occurences
				varID = p.substitution.newVar(i);
				for(int j = 0; j < occurences; j++){ // for each occurence
					c--; // move to next index 
					rule[rule[c]] = varID; // overwrite the rule part 
				}
			}
		}
	}
	
	/**
	 * Create a list with integers which flattens out all variables.
	 * @param rule Rule to be converted.
	 * @param cursor Start of the conversion.
	 * @return List with integers.
	 */
	public PrimIntArrayList deepcopy(int[] rule, int cursor){
		deepcopymode = true;
		deepcopycursor = cursor;
		array1.clear();
		int end = scanElement(rule, cursor, true, false);
		for(int i = deepcopycursor; i < end; i++) // add ints from last var to the end
			array1.addInt(rule[i]);
		deepcopymode = false; 
		return array1;
	}
	/**
	 * Scan over an element.
	 * @param rule Rule to be scanned.
	 * @param start Starting index.
	 * @param copy Whether to save the skipped integers somewhere.
	 * @param invar Whether the rule is inside a variable.
	 * @return The end index after the scan.
	 */
	public int scanElement(int[] rule, int start, boolean copy, boolean invar){
		int type = rule[start];
		if(type==PREDICATE){
			int x = start+3; // start of args
			int nrargs = rule[start+2];
			for(int i = 0; i < nrargs;i++) x = scanElement(rule,x,copy,invar); // scan over arguments
			return x;
		} else if(type==VARIABLE){
			int cur_id = rule[start+1];
			if(copy){
				if(invar&&cur_id == Integer.MIN_VALUE){
					cur_id = p.substitution.newVar(-1);
					rule[start+1] = cur_id;
				}
			}
			if(deepcopymode){ 
				for(int i = deepcopycursor; i < start; i++) // add ints from last var to this one
					array1.addInt(rule[i]);
				int[] value = p.substitution.value(cur_id); // get the var's value 
				if(value!=null){ // if it isn't null: 
					deepcopycursor = 0; // deep copy the value 
					int end = scanElement(value, 0, true, true);
					for(int i = deepcopycursor; i < end; i++) // add ints from last var to the end
						array1.addInt(value[i]);
				} else { // else add it as variable
					// TODO: een deep copy rule modus die ervor zorgt dat de vars netjes worden bijgehouden
					array1.addInt(IntHarvester.VARIABLE);
					array1.addInt(cur_id);
				}
				deepcopycursor = start+2;// move the deep copy cursor
			}
			return start+2;
		} else if(type==LIST){
			int x = start+2; // start of elems
			int nrElems = rule[start+1];
			for(int i = 0; i < nrElems;i++) x = scanElement(rule,x,copy,invar); // scan over elems
			if(nrElems>0)x = scanElement(rule,x,copy,invar);// also skip tail
			return x; 
		} else if(type==NUMBER){
			return start+3;
		} else if(type==PARENTHESIZED){
			int x = start+2; // start of args
			int nrElems = rule[start+1];
			for(int i = 0; i < nrElems;i++) x = scanElement(rule,x,copy,invar); // scan over arguments
			return x;
		} else if(type==CUT){ 
			return start+2;
		} else {
			return -1;
		}
	}
	/**
	 * Unify two terms.
	 * @param rule1 One of the rules.
	 * @param rule2 One of the rules.
	 * @param cursor1 Cursor of the first rule.
	 * @param cursor2 Cursor of the second rule. 
	 * @param invar1 Whether the first rule is inside a variable.
	 * @param invar2 Whether the second rule is inside a variable.
	 * @return Whether the terms unify.
	 */
	public boolean unifyTerm(int[] rule1, int[] rule2, int cursor1,int cursor2, boolean invar1, boolean invar2){
		if(p.trace){
			int[] ar1 = p.extractArray(rule1,cursor1,scanElement(rule1, cursor1, true, false));
			int[] ar2 = p.extractArray(rule2,cursor2,scanElement(rule2, cursor2, true, false));
			System.out.println(": unifying "+arToString(ar1, false)+" with "+arToString(ar2, false));
		}
		int type1 = rule1[cursor1];
		int type2 = rule2[cursor2];
		if(type1 == VARIABLE || type2==VARIABLE) return handleVars(rule1,rule2,cursor1,cursor2,invar1,invar2); 
		if(type1!=type2) return false; // types do not match
		else if(type1==NUMBER) return rule1[cursor1+1]==rule2[cursor2+1]&&rule1[cursor1+2]==rule2[cursor2+2]; // number must be the same
		else if(type1==PREDICATE||type1==PARENTHESIZED){
			if(type1==PREDICATE)
				if(rule1[cursor1+1]!=rule2[cursor2+1])
					return false; // predicate id's do not match
			return unifyArguments(rule1,rule2,cursor1,cursor2,invar1,invar2);
		} else if(type1==LIST) return unifyList2(rule1, rule2, cursor1+2, cursor2+2, rule1[cursor1+1], rule2[cursor2+1], rule1[cursor1+1], rule2[cursor2+1],invar1,invar2);
		return false;
	}
	/**
	 * Handle variables inside a rule.
	 * @param rule1 One of the rules.
	 * @param rule2 One of the rules.
	 * @param cursor1 Cursor of the first rule.
	 * @param cursor2 Cursor of the second rule. 
	 * @param invar1 Whether the first rule is inside a variable.
	 * @param invar2 Whether the second rule is inside a variable.
	 * @return Whether the situation unifies.
	 */
	public boolean handleVars(int[] rule1, int[] rule2, int cursor1, int cursor2, boolean invar1, boolean invar2){
		boolean isvar1 = rule1[cursor1] == VARIABLE;
		boolean isvar2 = rule2[cursor2] == VARIABLE;
		int id1 = isvar1?rule1[cursor1+1]:-1;
		int id2 = isvar2?rule2[cursor2+1]:-1; 
		if(!invar1&&isvar1&&id1==Integer.MIN_VALUE) return true; // don't care values while not inside a variable, so their binding won't matter
		if(!invar2&&isvar2&&id2==Integer.MIN_VALUE) return true; 
		if(invar1&&isvar1&&id1==Integer.MIN_VALUE){ // don't care values while inside a variable, these bindings might matter
			id1 = p.substitution.newVar(-1);
			rule1[cursor1+1] = id1;
		}
		if(invar2&&isvar2&&id2==Integer.MIN_VALUE){ // don't care values while inside a variable, these bindings might matter
			id2 = p.substitution.newVar(-1);
			rule2[cursor2+1] = id2;
		}
		if(isvar1&&!isvar2) return checkVar(id1,rule1,rule2,cursor1,cursor2,true,invar1,invar2); // one of them is not a var
		if(isvar2&&!isvar1) return checkVar(id2,rule1,rule2,cursor1,cursor2,false,invar1,invar2);// rule1 vars must already be instantiated
		if(id1 == id2) return true; // for A=A queries
		
		int[] v1 = p.substitution.value(id1);
		int[] v2 = p.substitution.value(id2);
		boolean isBound1 = v1 != null;
		boolean isBound2 = v2 != null;
		if(isBound1&&isBound2) return unifyTerm(v1, v2, 0, 0,true,true); // both have a value so look if those can be unified
		if(!isBound1&&isBound2) p.substitution.bind(id1, v2); // if either one is bound: bind the other as well
		if(isBound1&&!isBound2) p.substitution.bind(id2, v1);
		p.substitution.share(id1, id2); // share them
		return true;
	}
	/**
	 * If one value is a variable, and the other not, this unify is called to check whether the situation is unifiable. 
	 * @param id Variable id.
	 * @param rule1 One of the rules.
	 * @param rule2 One of the rules.
	 * @param cursor1 Cursor of the first rule.
	 * @param cursor2 Cursor of the second rule.
	 * @param firstIsVar Is rule1 the one with the variable?
	 * @param invar1 Whether the first rule is inside a variable.
	 * @param invar2 Whether the second rule is inside a variable.
	 * @return Whether the variable unifies with the other term.
	 */
	public boolean checkVar(int id, int[] rule1, int[] rule2, int cursor1, int cursor2, boolean firstIsVar, boolean invar1, boolean invar2){
		int[] v = p.substitution.value(id); // get the value of the variable
		if(v != null&&firstIsVar) return unifyTerm(v,rule2,0,cursor2,true,invar2); // if it is bound, check the value with the argument value
		if(v != null&&!firstIsVar) {
			boolean result = unifyTerm(rule1,v,cursor1,0,invar1,true);
			return result; // if it is bound, check the value with the argument value
		}
		// else: bind the variable
		int[] array = null; 
		if(!firstIsVar)array = p.extractArray(rule1, cursor1, scanElement(rule1, cursor1, true, invar1));
		else array = p.extractArray(rule2, cursor2, scanElement(rule2, cursor2, true, invar2));
		p.substitution.bind(id,array); // bind the variable
		return true;
	}
	/**
	 * Unify the arguments of two predicates.
	 * @param rule1 One of the rules.
	 * @param rule2 One of the rules.
	 * @param cursor1 Cursor of the first rule.
	 * @param cursor2 Cursor of the second rule.
	 * @param invar1 Whether the first rule is inside a variable.
	 * @param invar2 Whether the second rule is inside a variable.
	 * @return Whether all arguments unify.
	 */
	public boolean unifyArguments(int[] rule1, int[] rule2, int cursor1, int cursor2, boolean invar1, boolean invar2){
		int amount1 = rule1[cursor1]==PREDICATE?rule1[cursor1+2]:rule1[cursor1+1];
		int amount2 = rule2[cursor2]==PREDICATE?rule2[cursor2+2]:rule2[cursor2+1];
		if(amount1!=amount2) return false; // one is longer than the other
		cursor1 += rule1[cursor1]==PREDICATE?3:2; // bring cursors to start of first argument
		cursor2 += rule2[cursor2]==PREDICATE?3:2;
		for(int arg = 0; arg < amount1; arg++){
			if(!unifyTerm(rule1,rule2,cursor1,cursor2,invar1,invar2)){ // arguments do not match
				return false;
			} else {
				cursor1 = scanElement(rule1, cursor1, false, false); // scan over the argument TODO: you already scanned over them, reuse answer somewhere?
				cursor2 = scanElement(rule2, cursor2, false, false);
			}
		}
		return true;
	}

	/**
	 * Unify two lists.
	 * @param rule1 One of the rules.
	 * @param rule2 One of the rules.
	 * @param cursor1 Cursor of the first rule.
	 * @param cursor2 Cursor of the second rule.
	 * @param size1 Size of rule 1.
	 * @param size2 Size of rule 2.
	 * @param counter1 Elements to go of rule 1.
	 * @param counter2 Elements to go of rule 2.
	 * @param invar1 Whether the first rule is inside a variable.
	 * @param invar2 Whether the second rule is inside a variable.
	 * @return Whether they unify.
	 */
	public boolean unifyList2(int[] rule1, int[] rule2, int cursor1, int cursor2, int size1, int size2, int counter1, int counter2, boolean invar1, boolean invar2){ // cursors must be at the next element to check
		if(size1==0&&size2==0)return true;// both []
		if(size1==0&&size2>0||size2==0&&size1>0) return false; // one is [] other not, need this check because [] doesn't have tail
		while(counter1>0 && counter2>0){ // keep scanning through the elements
			if(!unifyTerm(rule1,rule2,cursor1,cursor2,invar1,invar2)) return false; // an element doesn't unify
			counter1--;
			counter2--;
			cursor1 = scanElement(rule1, cursor1, false, false); // move the cursors TODO same story as other todo
			cursor2 = scanElement(rule2, cursor2, false, false);
		}
		if(counter1==0&&counter2==0) return unifyTerm(rule1,rule2,cursor1,cursor2,invar1,invar2); // both hit tail at the same time
		if(counter1==0&&counter2>0) return unifyTail(rule1,rule2,cursor1,cursor2,counter2,size2,true,invar1,invar2); // rule1 vars are already instantiated
		if(counter1>0&&counter2==0) return unifyTail(rule1,rule2,cursor1,cursor2,counter1,size1,false,invar1,invar2); // rule2 vars are not always instantiated
		return true;
	}
	
	/**
	 * Unify the tails of two lists.
	 * @param rule1 One of the rules.
	 * @param rule2 One of the rules.
	 * @param cursor1 Cursor of the first rule.
	 * @param cursor2 Cursor of the second rule.
	 * @param counter Counter of nr. elements to go.
	 * @param size Size of the list that is not yet at the tail.
	 * @param callHitTail Whether rule1 (true) has hit the tail or rule2 (false).
	 * @param invar1 Whether the first rule is inside a variable.
	 * @param invar2 Whether the second rule is inside a variable.
	 * @return Whether they unify.
	 */
	public boolean unifyTail(int[] rule1, int[] rule2, int cursor1, int cursor2, int counter,int size, boolean callHitTail, boolean invar1, boolean invar2){
		int[] tailrule = callHitTail?rule1:rule2;
		int tailcursor = callHitTail?cursor1:cursor2;
		if(tailrule[tailcursor]==VARIABLE){
			int id = tailrule[tailcursor+1];
			if(callHitTail&&!invar1&&id==Integer.MIN_VALUE)return true;
			if(!callHitTail&&!invar2&&id==Integer.MIN_VALUE)return true;
			if(callHitTail&&invar1&&id==Integer.MIN_VALUE){
				id = p.substitution.newVar(-1);
				rule1[cursor1+1] = id;
			}
			if(!callHitTail&&invar2&&id==Integer.MIN_VALUE){
				id = p.substitution.newVar(-1);
				rule2[cursor2+1] = id;
			}

			int[] varval = p.substitution.value(id);
			if(varval==null) // tail is an unbound variable
				return recordTail(id,(callHitTail?rule2:rule1),(callHitTail?cursor2:cursor1),counter,callHitTail?invar2:invar1); // record the rest of the second list in the tail
			else if(varval[0]!=LIST) return false;
			else if(callHitTail)
				return unifyList2(varval,rule2,2,cursor2,varval[1],size,varval[1],counter,true,invar2);
			else return unifyList2(rule1,varval,cursor1,2,size,varval[1],counter,varval[1],invar1,true);
		} else if(tailrule[tailcursor]!=LIST) return false;
		else if(callHitTail){
			return unifyList2(rule1,rule2,cursor1+2,cursor2,rule1[cursor1+1],size,rule1[cursor1+1],counter,invar1,invar2); // the one with instantiated vars must always come first
		} else {
			return unifyList2(rule1,rule2,cursor1,cursor2+2,size,rule2[cursor2+1],counter,rule2[cursor2+1],invar1,invar2);
		}
	}
	
	/**
	 * Record the values of a given list from a given point onwards and put those in a variable.
	 * @param varID Id of the to-be-bound variable.
	 * @param rule The rule in which the list occurs.
	 * @param cursor The cursor in the rule. 
	 * @param counter Amount of elements to record.
	 * @param invar Check whether you are in a variable (has to do with possible anonymous variables).
	 * @return Whether record action was successful.
	 */
	public boolean recordTail(int varID, int[] rule, int cursor, int counter, boolean invar){
		if(p.trace)System.out.println(": recording list to put in var with global id " +varID+" #elements to add: "+counter); 
		int newsize = counter;
		int b = cursor;
		while(counter >= 0){ // ">=" because you always have a tail at this point, if not: then the var would be at the same time encountered as the end of the list and that is already handled in unifyList
			b = scanElement(rule, b, true, invar);
			counter--;
		}
		int copySize = b-cursor;
		int[] prim = new int[copySize+2];
		prim[0] = LIST;
		prim[1] = newsize;
		int c = 2;
		for(int x = 0; x < copySize; x++){
			prim[c] = rule[cursor+x];
			c++;
		}
		p.substitution.bind(varID, prim);
		return true;
	}
	
	/**
	 * Convert an array to a string.
	 * @param The array to be converted.
	 * @param renamevars Whether to rename variables.
	 * @return The string representation.
	 */
	public String arToString(int[] ar, boolean renamevars){
		IDtoID.clear();
		if(ar == null) return "undefined";
		String head = elemToString(ar, 0,renamevars);
		int c = scanElement(ar, 0, false, false);
		if(c < ar.length){
			head += " :- ";
			while(c < ar.length){
				head += elemToString(ar, c,renamevars)+",";
				c = scanElement(ar, c, false, false);
			}
			head = head.substring(0,head.length()-1)+".";
		}
		return head;
	}
	
	/**
	 * Convert a integer rule to a string.
	 * @param rule The rule to be converted.
	 * @return The string representation.
	 */
	public String ruleToString(int[] rule){
		IDtoID.clear();
		String r = "";
		r += elemToString(rule,0,true);
		int c = scanElement(rule, 0, false, false);
		if(rule[rule.length-1]>1){
			r += " :- ";
			for(int i = 1; i < rule[rule.length-1]; i++){
				r += elemToString(rule, c,true)+",";
				c = scanElement(rule, c, false, false);
			}
			r = r.substring(0,r.length()-1)+".";
		}
		return r;
	}

	public boolean showVarIDs = true;
	PrimIntKeyHashMap<Integer> IDtoID = new PrimIntKeyHashMap<Integer>(PrimIntKeyHashMap.INT);
	/**
	 * Output an element.
	 * @param value Array where the element is part of.
	 * @param cursor Index where the element starts.
	 * @param renamevars Whether to rename variables.
	 * @return String representation.
	 */
	public String elemToString(int[] value, int cursor, boolean renamevars){
		if(value == null) return "_";
		int type = value[cursor];
		if(type==CUT) return "!";
		else if(type==VARIABLE){ 
			if(showVarIDs){
				if(renamevars){
					 int id = IDtoID.getInt(value[cursor+1]);
					 if(id<0){
						 id = IDtoID.size();
						 IDtoID.putInt(value[cursor+1], id);
					 }
					 return ""+ (char)((int)'A'+id);
				} else return value[cursor+1]!=Integer.MIN_VALUE?"{V"+value[cursor+1]+":"+elemToString(p.substitution.value(value[cursor+1]),0,renamevars)+"}":"_";
			}
			else return elemToString(p.substitution.value(value[cursor+1]), 0, renamevars);
		}
		else if(type==NUMBER){
			long l1 = ((long)value[cursor+1]<<32)>>>32;
			long l2 = ((long)value[cursor+2]<<32)>>>32;
			double d = Double.longBitsToDouble((l1<<32)|l2);
			return ""+d;
		} else if(type==PREDICATE){
			int x = cursor+3; // start of args
			String r = p.strStorage.getString(value[cursor+1])+(value[cursor+2]>0?"(":"");
			for(int i = 0; i < value[cursor+2];i++){
				r += elemToString(value,x,renamevars); 
				x = scanElement(value,x,false,false); // scan over arguments 
				if(i < value[cursor+2]-1) r += ",";
			}
			return r+(value[cursor+2]>0?")":"");
		} else if(type==PARENTHESIZED){
			int x = cursor+2; // start of args
			String r = "(";
			for(int i = 0; i < value[cursor+1];i++){
				r += elemToString(value, x,renamevars);
				x = scanElement(value,x,false,false); // scan over arguments
				if(i < value[cursor+1]-1) r += ",";
			}
			return r+")";
		} else if(type==LIST){
			int x = cursor+2; // start of elems
			String r = "[";
			for(int i = 0; i < value[cursor+1];i++){
				r += elemToString(value, x,renamevars);
				x = scanElement(value,x,false,false); // scan over elems
				if(i < value[cursor+1]-1) r += ",";
			}
			if(value[cursor+1]>0){
				r += "|"+elemToString(value, x,renamevars);
				x = scanElement(value,x,false,false);// also skip tail
			}
			return r+"]"; 
		} else {
			return "";
		}
	}
}
