package aplprolog.prolog;

import aplprolog.datastructures.PrimIntKeyHashMap;
/*
 * Nieuw idee:
 * - Unifying call p(X,Y) met p(3,4) genereert nu arrays voor X [number,3a,3b] en Y [number,4a,4b]. 
 * - Ipv nieuwe arrays kun je ook zeggen: verwijs naar de rule en geef twee indexen.
 * - Vooral als de calls ground zijn (i.e. veel bij 2apl en 2opl) scheelt dit veel: een call verandert nooit, ook niet zijn variabelen.
 * - Rule variabelen kunnen wel veranderen, dus let daar op. Met rule var overhead integers kun je dit echter makkelijk detecteren.
 * - Ook lijsten zijn tricky, als je een deel van een lijst moet recorden dan kun je niet naar 1 array met a en b verwijzen.
 * 
 */
/**
 * Substitution structure. Keeps track of which variables are bound with what. 
 * @author Bas Testerink, Utrecht University, The Netherlands
 *
 */
public class Substitution {
	Prolog p;
	PrimIntKeyHashMap<Integer> idmap = new PrimIntKeyHashMap<Integer>(PrimIntKeyHashMap.INT,512);
	int[] sharedHeads = new int[512];
	int[] sharedLists = new int[1024];
	int[][] values = new int[512][];
	public int id = 0;
	int headsFirst = 1; // 0 will be used as "not instantiated"
	int headsLast = 2; 
	int listsFirst = 2; 
	int listsLast = 4; 
	
	public Substitution(Prolog p){
		this.p = p;
		sharedHeads[headsFirst] = headsLast;
		sharedLists[listsFirst] = listsLast;
	}
	/**
	 * Create a new variable.
	 * @param localID Old, local id.
	 * @return New id.
	 */
	public int newVar(int localID){ // TODO een p.lastIndex voor de index v/d value, zodat bij bind je direct kunt assignen ?
		id++;
		int index = nextHeadsIndex(); // create a new index (shared lists are instantiated when needed)
		idmap.putInt(id, index);
		p.history.addInt(id);
		p.history.addInt(p.MAKEVAR);
		if(p.trace)System.out.println(": make variable for local id "+localID+", the variable gets global id: "+id); 
		return id;
	}
	/**
	 * Remake a variable.
	 * @param id Id that is to be remade.
	 * @return New id to replace the old one.
	 */
	public int remakeVar(int id){
		int index = nextHeadsIndex(); // create a new index (shared lists are instantiated when needed)
		idmap.putInt(id, index);
		p.history.addInt(id);
		p.history.addInt(p.MAKEVAR);
		return index;
	}
	/**
	 * Get next free heads space.
	 * @return Free heads space index.
	 */
	public int nextHeadsIndex(){
		int r = headsFirst; 
		if(headsFirst == headsLast){ // no more free holes in the heads array
			headsFirst++; // so move them onwards
			headsLast++; 
			if(headsFirst >= sharedHeads.length) resizeSH();
		} else {
			headsFirst = sharedHeads[headsFirst]; // else set the next free space
		}
		sharedHeads[r] = 0; // undo the next pointer
		return r;
	}
	
	/**
	 * Stretch the shared heads data structure.
	 */
	public void resizeSH(){
		int[] newSH = new int[sharedHeads.length*2];
		for(int i = 0; i < sharedHeads.length; i++) newSH[i] = sharedHeads[i];
		int[][] newValues = new int[values.length*2][];
		for(int i = 0; i < values.length; i++) newValues[i] = values[i];
		sharedHeads = newSH;
		values = newValues;
	}
	
	/**
	 * Delete a variable.
	 * @param id Id of the variable to be deleted.
	 */
	public void delete(int id){
		int index = idmap.removeInt(id);
		appendFreeHeadsSpace(index);
		values[index] = null;
		if(p.trace)System.out.println(": remove variable with global id "+id);
	}
	
	/**
	 * Add free space to the shared lists data structure.
	 * @param index Index to be added.
	 */
	public void appendFreeHeadsSpace(int index){
		sharedHeads[index] = headsFirst; // set the index's next free space to the current free space
		headsFirst = index; // set the first index to the new index
	}
	
	/**
	 * Share two variables.
	 * @param a One of the variables.
	 * @param b One of the variables.
	 */
	public void share(int a, int b){
		int indexA = idmap.getInt(a);
		int indexB = idmap.getInt(b);
		if(indexA<0)indexA=remakeVar(a);
		if(indexB<0)indexB=remakeVar(b);
		addToShared(indexA,indexB);
		addToShared(indexB,indexA);
		p.history.addInt(a);
		p.history.addInt(b);
		p.history.addInt(p.SHAREVAR);
		if(p.trace)System.out.println(": share the variables with global id: "+a+" and "+b);
	}
	
	/**
	 * Add a to b's shared list.
	 * @param a One of the variables.
	 * @param b One of the variables.
	 */
	public void addToShared(int a, int b){ 
		int next = nextListsIndex(sharedHeads[a],b); // space for the entry
		sharedHeads[a] = next; // the first entry index
	}
	/**
	 * Support function. Add a value to a shared list.
	 * @param next Current last added item to the shared list.
	 * @param value New value to be added.
	 * @return Index of the new value.
	 */
	public int nextListsIndex(int next, int value){
		int r = listsFirst; 
		if(listsFirst == listsLast){ // no more free holes in the heads array
			listsFirst += 2; // so move them onwards
			listsLast += 2; 
			if(listsFirst >= sharedLists.length) resizeSL();
		} else {
			listsFirst = sharedLists[listsFirst]; // else set the next free space
		}
		sharedLists[r] = next; // set the data
		sharedLists[r+1] = value;
		return r;
	}
	
	/**
	 * Support method, stretch shared lists data structure.
	 */
	public void resizeSL(){
		int[] newSL = new int[sharedLists.length*2];
		for(int i = 0; i < sharedLists.length; i++) newSL[i] = sharedLists[i];
		sharedLists = newSL;
	}
	/**
	 * Unshare two variables.
	 * @param a One of the two variables.
	 * @param b One of the two variables.
	 */
	public void unshare(int a, int b){ // unshare happens only to the last added share, so is always the last of the shared list
		int indexA = idmap.getInt(a);
		int indexB = idmap.getInt(b);
		removeLastShared(indexA,sharedHeads[indexA]);
		removeLastShared(indexB,sharedHeads[indexB]);
		if(p.trace)System.out.println(": UNshare the variables with global id: "+a+" and "+b);
	}
	
	/**
	 * Remove the last added shared variable.
	 * @param headIndex Index in the bookkeeping datastructures.
	 * @param entryIndex Index in the bookkeeping datastructures.
	 */
	public void removeLastShared(int headIndex, int entryIndex){  
		int next = sharedLists[entryIndex]; // the new first entry
		sharedLists[entryIndex] = listsFirst; // append the entry to the free spaces list
		listsFirst = entryIndex; // set the new first pointer to the removed entry
		sharedHeads[headIndex] = next; // move the head to the next entry 
	}
	
	/**
	 * Bind a variable.
	 * @param id Id of the variable.
	 * @param value Value to bind the variable with.
	 */
	public void bind(int id, int[] value){
		int index = idmap.getInt(id); // get the index
		if(index<0)index=remakeVar(id);
		sharedBind(index,value);
		p.history.addInt(id);
		p.history.addInt(p.BINDVAR); 
		if(p.trace)System.out.println(": bind the variable with global id "+id+" with the value "+p.arStr(value)+" interpreted: "+p.harvester.arToString(value,false));
	}
	
	/**
	 * Bind a chain of shared variables.
	 * @param index Some variable in the chain.
	 * @param value Value for the variables.
	 */
	public void sharedBind(int index, int[] value){
		if(values[index]==null){
			values[index] = value;
			int cursor = sharedHeads[index];
			while(cursor!=0){ // for each shared variable
				sharedBind(sharedLists[cursor+1],value); // bind the shared variable
				cursor = sharedLists[cursor]; // move the cursor
			}
		}
	}
	
	/**
	 * Free a variable.
	 * @param id Id of the variable.
	 */
	public void free(int id){
		sharedFree(idmap.getInt(id));
		if(p.trace)System.out.println(": free the variable with global id "+id);
	}
	
	/**
	 * Free a chain of shared variables.
	 * @param index Some variable in the chain.
	 */
	public void sharedFree(int index){
		if(values[index]!=null){
			values[index] = null;
			int cursor = sharedHeads[index];
			while(cursor!=0){ // for each shared variable
				sharedFree(sharedLists[cursor+1]); // free the shared variable
				cursor = sharedLists[cursor]; // move the cursor
			}
		}
	}
	
	/**
	 * Clear the entire substitution.
	 */
	public void clear(){
		for(int i = 0; i < values.length; i++){
			values[i] = null;
			sharedHeads[i] = 0;
		}
		for(int i = 0; i < sharedLists.length; i++)sharedLists[i] = 0;
		idmap.clear();
		id = 0;
		headsFirst = 1; // 0 will be used as "not instantiated"
		headsLast = 2; 
		listsFirst = 2; 
		listsLast = 4; 
		sharedHeads[headsFirst] = headsLast;
		sharedLists[listsFirst] = listsLast;
	}
	
	/**
	 * Value that belongs to a variable.
	 * @param id Id of the variable.
	 * @return The value (null if not bound).
	 */
	public int[] value(int id){
		int index = idmap.getInt(id);
		if(index<=0)return null;
		return values[index];
	}
}
