package aplprolog.prolog.builtins;
import aplprolog.prolog.*;
/**
 * Built in functions for arithmetical actions.
 * 
 * @author Bas Testerink, Utrecht University, The Netherlands
 *
 */
public class Arithmetics {
	Prolog p;
	BuiltIn bi;

	public Arithmetics(BuiltIn bi){
		this.p = bi.p;
		this.bi = bi;
	}

	public int arithcursor = 0;
	/**
	 * Check whether you can successfully perform the arithmetical call.
	 * @param call The arithmetical call.
	 * @return Whether the arithmetical call could be executed.
	 */
	public boolean arithmetic(int[] call){
		if(call[arithcursor]!=IntHarvester.VARIABLE && call[arithcursor]!=IntHarvester.NUMBER){
			if(p.trace)System.out.println(": left-hand of the arithmetical expression is no unbound variable or number");
			return false; // first elem must be a variable or number
		} if(call[arithcursor]==IntHarvester.VARIABLE){
			int[] varv = p.substitution.value(call[arithcursor+1]);
			if(varv!=null){
				if(varv[0]!=IntHarvester.NUMBER){
					if(p.trace)System.out.println(": left-hand of the arithmetical expression is no unbound variable or number");
					return false;
				}
			}
		}
		double d = evaluate(call); // is d a variable or a double
		double eval = evaluate(call);
		if(Double.isNaN(eval)){
			if(p.trace)System.out.println(": not a proper arithmetical expression");
			return false;
		}
		if(Double.isNaN(d)){ // d is a variable
			long l = Double.doubleToLongBits(eval);
			int a = (int)((l>>>32));
			int b = (int)((l<<32)>>>32);
			p.substitution.bind(call[4], new int[]{IntHarvester.NUMBER,a,b});
		} else return d == eval;
		return true;
	}
	/**
	 * Various evaluation methods.
	 * @param call Call with the built-in predicate/function.
	 * @return Resulting number.
	 */
	public double evaluate(int[] call){
		if(call[arithcursor]==IntHarvester.NUMBER){
			long l1 = ((long)call[arithcursor+1]<<32)>>>32;
			long l2 = ((long)call[arithcursor+2]<<32)>>>32;
			arithcursor += 3;
			return Double.longBitsToDouble((l1<<32)|l2);
		} else if(call[arithcursor]==IntHarvester.VARIABLE){
			if(p.substitution.value(call[arithcursor+1])==null){
				arithcursor += 2;
				return Double.NaN; // not a proper instatiated var
			}
			int c = arithcursor; // remind the current position
			arithcursor = 0; // position for evaluating var value
			double d = evaluate(p.substitution.value(call[c+1])); // get the evaluation of the variable
			arithcursor = c+2; // return the arithmetic cursor
			return d; // return the value
		} else if(call[arithcursor]==IntHarvester.PARENTHESIZED){
			arithcursor+=2; // ignore brackets
			return  evaluate(call);
		} else if(call[arithcursor]==IntHarvester.PREDICATE){
			int operator = call[arithcursor+1];
			arithcursor += 3;
			double a = Double.NaN;
			double b = Double.NaN;
			switch(operator-bi.range){
			case BuiltIn.ADD: a = evaluate(call); b = evaluate(call); return a+b;
			case BuiltIn.SUBTRACT:
				if(call[arithcursor-1]==2){a = evaluate(call); b = evaluate(call); return a-b;}
				else if(call[arithcursor-1]==1){ a = evaluate(call); return -a;}
			case BuiltIn.MULTIPLY: a = evaluate(call); b = evaluate(call); return a*b;
			case BuiltIn.POW: a = evaluate(call); b = evaluate(call); return Math.pow(a, b);
			case BuiltIn.STARPOW: a = evaluate(call); b = evaluate(call); return Math.pow(a, b);
			case BuiltIn.DIVIDE: a = evaluate(call); b = evaluate(call); return a/b;
			case BuiltIn.INTDIVIDE: 	a = evaluate(call); b = evaluate(call);
			if(!areInts(a,b)){
				if(p.trace)System.out.println(": no can do: integer division on a double");
				return Double.NaN;
			} else return (int)(a/b); 
			case BuiltIn.REMAINDER: a = evaluate(call); b = evaluate(call); // difference mod and rem: one is truncate, other floor
			if(!areInts(a,b)){
				if(p.trace)System.out.println(": no can do: remainder on a double");
				return Double.NaN;
			} else return a-(((int)(a/b))*b);
			case BuiltIn.MODULO: a = evaluate(call); b = evaluate(call);
			if(!areInts(a,b)){
				if(p.trace)System.out.println(": no can do: remainder on a double");
				return Double.NaN;
			} else return a-(Math.floor(a/b)*b);
			case BuiltIn.ABS: a = evaluate(call); return Math.abs(a);
			case BuiltIn.SIGN: a = evaluate(call); return Math.signum(a);
			case BuiltIn.INTEGERPART: a = evaluate(call); return a - (a-(a>0?Math.floor(a):Math.ceil(a)));
			case BuiltIn.FRACPART: a = evaluate(call); return a-(a>0?Math.floor(a):Math.ceil(a));
			case BuiltIn.FLOAT: a = evaluate(call); return a; // everything is float
			case BuiltIn.FLOOR: a = evaluate(call); return Math.floor(a);
			case BuiltIn.TRUNCATE: a = evaluate(call); return a - (a-(a>0?Math.floor(a):Math.ceil(a)));
			case BuiltIn.CEIL: a = evaluate(call); return Math.ceil(a);
			case BuiltIn.CEILING: a = evaluate(call); return Math.ceil(a);
			case BuiltIn.MAX: a = evaluate(call); b = evaluate(call); return Math.max(a, b);
			case BuiltIn.MIN: a = evaluate(call); b = evaluate(call); return Math.min(a, b);
			case BuiltIn.INTEGER: a = evaluate(call); return Math.round(a);
			case BuiltIn.ROUND: a = evaluate(call); return Math.round(a);
			case BuiltIn.SHIFTLEFT: a = evaluate(call); b = evaluate(call);
			if(!areInts(a,b)){
				if(p.trace)System.out.println(": no can do: bitshift on doubles");
				return Double.NaN;
			} else return (double)(((long)a)<<((long)b));
			case BuiltIn.SHIFTRIGHT: a = evaluate(call); b = evaluate(call);
			if(!areInts(a,b)){
				if(p.trace)System.out.println(": no can do: bitshift on doubles");
				return Double.NaN;
			} else return (double)(((long)a)>>>((long)b));
			case BuiltIn.OR: a = evaluate(call); b = evaluate(call);
			if(!areInts(a,b)){
				if(p.trace)System.out.println(": no can do: bitshift on doubles");
				return Double.NaN;
			} else return (double)(((long)a)|((long)b));
			case BuiltIn.AND: a = evaluate(call); b = evaluate(call);
			if(!areInts(a,b)){
				if(p.trace)System.out.println(": no can do: bitshift on doubles");
				return Double.NaN;
			} else return (double)(((long)a)&((long)b));
			case BuiltIn.XOR: a = evaluate(call); b = evaluate(call);
			if(!areInts(a,b)){
				if(p.trace)System.out.println(": no can do: bitshift on doubles");
				return Double.NaN;
			} else return (double)(((long)a)^((long)b));
			case BuiltIn.COMPLEMENT: a = evaluate(call);
			if(!isInt(a)){
				if(p.trace)System.out.println(": no can do: bitshift on doubles");
				return Double.NaN;
			} else return (double)~((long)a);
			case BuiltIn.SQRT: a = evaluate(call); return Math.sqrt(a);
			case BuiltIn.TAN: a = evaluate(call); return Math.tan(a);
			case BuiltIn.COS: a = evaluate(call); return Math.cos(a);
			case BuiltIn.SIN: a = evaluate(call); return Math.sin(a);
			case BuiltIn.ATAN: a = evaluate(call); return Math.atan(a);
			case BuiltIn.ACOS: a = evaluate(call); return Math.acos(a);
			case BuiltIn.ASIN: a = evaluate(call); return Math.asin(a);
			case BuiltIn.LOG: a = evaluate(call); return Math.log(a);
			case BuiltIn.LOG10: a = evaluate(call); return Math.log10(a);
			case BuiltIn.EXP: a = evaluate(call); return Math.exp(a);
			case BuiltIn.E: return Math.E;
			case BuiltIn.PI: return Math.PI;
			default: 
				if(p.trace)System.out.println(": operator/function "+p.strStorage.getString(operator)+" doesn't exist"); 
				return Double.NaN;
			}
		} else{
			return Double.NaN;
		}
	}

	/**
	 * Check whether a and b are integers.
	 * @param a A number.
	 * @param b A number.
	 * @return Whether a and b are integers.
	 */
	public boolean areInts(double a, double b){
		return isInt(a)&&isInt(b);
	}
	/**
	 * Check whether a is an integer.
	 * @param a Number to be checked.
	 * @return Whether a is an integer.
	 */
	public boolean isInt(double a){
		return a - Math.round(a) == 0;
	}
	/**
	 * Implementation of =:=. 
	 * @param call The call to the operator.
	 */
	public boolean arithmeticEQ(int[] call){ 
		double a = evaluate(call); // is d a variable or a double
		double b = evaluate(call);
		if(Double.isNaN(a)||Double.isNaN(b)){
			if(p.trace)System.out.println(": not a proper arithmetical expression");
			return false;
		}
		return a == b;
	}
}
