package aplprolog.prolog.builtins;

import aplprolog.datastructures.PrimIntArrayList;
import aplprolog.datastructures.PrimIntKeyHashMap;
import aplprolog.prolog.BuiltIn;
import aplprolog.prolog.IntHarvester;
import aplprolog.prolog.Prolog;
/**
 * Everything that is related to modification of the rulebase at runtime.
 * 
 * @author Bas Testerink, Utrecht University, The Netherlands
 *
 */
public class RulebaseModifications {
	BuiltIn bi;
	public PrimIntArrayList array = new PrimIntArrayList(true), splits = new PrimIntArrayList(true), changeLog = new PrimIntArrayList(false);
	PrimIntKeyHashMap<Integer> vars = new PrimIntKeyHashMap<Integer>(PrimIntKeyHashMap.INT);
	PrimIntKeyHashMap<PrimIntArrayList> varIndices = new PrimIntKeyHashMap<PrimIntArrayList>(PrimIntKeyHashMap.OBJ);
	int cursor = 0;
	int bodyparts = 0;
	int[] call = null;
	public boolean addZeroSplit = false; // adds a zerosplit when copying (for 2OPL), makes a term a query
	public boolean logModifications = false; // keep track of the modifications to the belief base (for 2OPL regimentation)

	public RulebaseModifications(BuiltIn bi){
		this.bi = bi;
	}

	/**
	 * Call for assertion based on a query. (the assert(A/Z)/1 built-in predicate functionality).
	 * @param ar The call.
	 * @param assertA Whether to assert at the front or at the tail.
	 * @return Success of the predicate.
	 */
	public boolean queryAssert(int[] ar, boolean assertA, boolean unique, boolean failAtUnique){
		// First search the correct term inside the array to be asserted
		call = ar;
		if(!getToTheArgument(3,true))return false; // no good argument

		// obtain a deep copy of the term and check the validity of the argument
		int[] copy = null;
		if(bodyparts==1&&call[cursor]==IntHarvester.PREDICATE&&(call[cursor+1]!=bi.IC_RULEIMPLY||call[cursor+2]!=2)){ // assert a simple fact
			copy = deepcopy(call, cursor, call.length); // the call ends always with the last argument, in this case the fact to be added
		} else if(call[cursor]==IntHarvester.PREDICATE&&call[cursor+1]==bi.IC_RULEIMPLY&&call[cursor+2]==2){ // assert a rule (starts with :-/2)
			copy = deepcopy(call,cursor+3,call.length); // skip over the :-/2 part
		} else return false; // no proper arguments  

		// check for storage space
		PrimIntKeyHashMap<PrimIntArrayList> nameBase = bi.p.rulebase.getObj(copy[1]); // add name rules
		if(nameBase == null){
			nameBase = new PrimIntKeyHashMap<PrimIntArrayList>(PrimIntKeyHashMap.OBJ);
			bi.p.rulebase.putObj(copy[1], nameBase);
		}
		PrimIntArrayList arityBase = nameBase.getObj(copy[2]); // add arity rules
		if(arityBase == null){
			arityBase = new PrimIntArrayList(false);
			nameBase.putObj(copy[2], arityBase);
		}
		boolean exists = false;
		if(unique){ // unique assert: do not assert if the rule/fact already exists
			for(int x = 0; x < arityBase.pointer; x++){
				int[] rule = arityBase.arraydata[x];
				if(rule[0]!=Prolog.TOBEDELETED&&rule.length == copy.length){ // fail fast check
					// set the proper variabe values (can be changed by IntHarvester's buildLocalVars())
					int c = rule.length-rule[rule.length-1]-2; // init as var cursor
					int nrVars = rule[c]; // the amount of local variables 
					int occurences = 0; // occurence counter
					if(nrVars > 0){// replace the vars
						for(int i = 0; i < nrVars; i++){ // for each variable
							c--; // move cursor to the variable occurence counter
							occurences = rule[c]; // amount of occurences
							for(int j = 0; j < occurences; j++){ // for each occurence
								c--; // move to next index 
								rule[rule[c]] = i; // overwrite the rule part 
							}
						}
					} // now the copy and the rule are equal if they would be equal after parsing
					boolean match = true;
					for(int i = 0; i < copy.length; i++) // see if the rule is unequal to the copy
						if(copy[i]!=rule[i]){
							match = false;
							i = copy.length;
						}
					if(match){ // if the rule is equal then the copy already exists
						exists = true;
						x = arityBase.pointer;
					}
				}
			}
		}
		// add the rule
		if((unique&&!exists)||!unique){ // if not unique: always assert, if unique: only assert if the copy does not already exist. Success guaranteed.
			if(assertA) arityBase.addAr(0,copy);
			else arityBase.addAr(copy);
			bi.p.rulesModified = true; 
			if(logModifications){ // note that this is quite heavy and one should avoid regimented 2OPL organizations
				addZeroSplit = true;
				int[] callCopy = deepcopy(call, 0, call.length);
				addZeroSplit = false;
				callCopy[1] = bi.IC_RETRACT; // it is immediately applicable as a query to undo this assertion
				changeLog.addAr(callCopy); 
			}
		}
		if(unique&&exists&&failAtUnique) return false;
		return true;
	}
	/**
	 * Built-in function for retract/1 and retractall/1. 
	 * @param ar The call.
	 * @param retractAll Whether the call is retract or retractall.
	 * @return Result of the predicate.
	 */
	public boolean queryRetract(int[] ar, boolean retractAll){
		call = ar;
		if(!getToTheArgument(3,true))return false; // no good argumentb
		// Get the correct name
		int name = call[cursor+1];
		if(name == bi.IC_RULEIMPLY){ // skip :-/2
			cursor += 3;
			name = call[cursor+1];
			bodyparts++; // correct bodyparts counter, last parenthesis is taken which would mean that ((a:-b)) is seen as (:-(a,b)) so 1 bodypart
		}

		// check whether entries exist
		PrimIntKeyHashMap<PrimIntArrayList> nameBase = bi.p.rulebase.getObj(name);
		if(nameBase == null)return retractAll; // no predicates that share the name
		PrimIntArrayList arityBase = nameBase.getObj(call[cursor+2]);
		if(arityBase == null)return retractAll; // no predicates with the same name and arity 
		for(int i = 0; i < arityBase.pointer; i++){ // TODO: SWI-Prolog retract fifo, but lafo is faster. Edit there for to lafo?
			int[] rule = arityBase.arraydata[i]; 
			if(rule[0]!=Prolog.TOBEDELETED&&rule[rule.length-1]==bodyparts){ // amount of bodyparts must match
				bi.p.harvester.buildLocalVars(rule); // amount of vars will be 0 in most applications (2APL and 2OPL cannot assert unground facts by definition)
				int partcursor = cursor; 
				int rulecursor = 0;
				boolean match = true;
				for(int part = 0; part < bodyparts; part++){ // unify each bodypart
					if(!bi.p.harvester.unifyTerm(call, rule, partcursor, rulecursor, false, false)){
						bi.p.backtrack(false);
						match = false;
						part = bodyparts; 
					}
					rulecursor = bi.p.harvester.scanElement(rule, rulecursor, false, false); // move the cursor
					partcursor = bi.p.harvester.scanElement(call, partcursor, false, false); // move the cursor
				}
				if(match){ 
					// remove the rule
					if(logModifications){ // note that this is quite heavy and one should avoid regimented 2OPL organizations
						addZeroSplit = true;
						int[] callCopy = deepcopy(call, 0, call.length);
						addZeroSplit = false;
						callCopy[1] = bi.IC_ASSERTZ; // it is immediately applicable as a query to undo this assertion
						changeLog.addAr(callCopy);
					}
					arityBase.arraydata[i][0] = Prolog.TOBEDELETED; // mark for deletion
					bi.p.rulesModified = true;
					if(retractAll) bi.p.backtrack(false);
					else i = arityBase.pointer;
				}
			}
		}
		return true;
	}

	/**
	 * Update a single number or atom name.
	 * @param ar The call.
	 * @return Whether update was successful.
	 */
	public boolean atomicUpdate(int[] ar){
		call = ar;
		int intA = 0;
		int intB = 0;
		int newPred = 0;
		int argNr = 0;
		if(!getToTheArgument(3,false))return false; // no good argument
		boolean updatePredName = call[cursor]==IntHarvester.PREDICATE;  // first argument is the new value
		if(updatePredName)newPred = call[cursor+1];
		else {
			intA = call[cursor+1];
			intB = call[cursor+2];
		}
		cursor = bi.p.harvester.scanElement(ar, 3, false, false);
		call = ar;
		if(!getToTheArgument(cursor, false)) return false;
		if(call[cursor]!=IntHarvester.NUMBER) return false; // second arg is argnr
		long l1 = ((long)call[cursor+1]<<32)>>>32;
		long l2 = ((long)call[cursor+2]<<32)>>>32;
		argNr = (int)Double.longBitsToDouble((l1<<32)|l2);
		call = ar;
		if(!getToTheArgument(bi.p.harvester.scanElement(ar, cursor, false, false), true)) return false;
		int name = ar[cursor+1];
		// check whether entries exist
		PrimIntKeyHashMap<PrimIntArrayList> nameBase = bi.p.rulebase.getObj(name);
		if(nameBase == null)return false; // no predicates that share the name
		PrimIntArrayList arityBase = nameBase.getObj(call[cursor+2]);
		if(arityBase == null)return false; // no predicates with the same name and arity 
		boolean match = false;
		for(int i = 0; i < arityBase.pointer; i++){ // TODO: SWI-Prolog retract fifo, but lafo is faster. Edit there for to lafo?
			int[] rule = arityBase.arraydata[i]; 
			if(rule[rule.length-1]==1){ // amount of bodyparts must be 0
				bi.p.harvester.buildLocalVars(rule); // amount of vars will be 0 in most applications (2APL and 2OPL cannot assert unground facts by definition)
				if(!bi.p.harvester.unifyTerm(call, rule, cursor, 0, false, false)){
					bi.p.backtrack(false);
				} else {  // update the value
					match = true;
					int argcursor = 3;
					match = true;
					for(int arg = 1; arg < argNr; arg++)
						argcursor = bi.p.harvester.scanElement(rule, argcursor, false, false);
					if(updatePredName)rule[argcursor+1] = newPred;
					else {
						rule[argcursor+1] = intA;
						rule[argcursor+2] = intB;
					}
					i = arityBase.pointer;
				}
			}
		}
		return match;
	}
	/**
	 * Go through the parenthesis and variable contents of a retract/assert argument. I.e.: "A=f(a),B=(((A))),assert(((B)))." asserts f(a).
	 * @return Whether the argument is legal.
	 */
	public boolean getToTheArgument(int startCursor, boolean predicateOnly){
		cursor = startCursor;
		bodyparts = 1; // 1 = fact, or x :- y, 2 = body with more than one subgoal.
		while(call[cursor] == IntHarvester.VARIABLE || // while the value is inside a variable
				call[cursor]==IntHarvester.PARENTHESIZED){ // or if the value is wrapped in parenthesis
			if(call[cursor] == IntHarvester.VARIABLE){
				call = bi.p.substitution.value(call[cursor+1]);
				cursor = 0;
				if(call == null) return false; // argument was unbound variable
			} else{
				bodyparts = call[cursor+1]; // amount of subgoals
				cursor += 2; // move through parenthesis
			}
		}
		if(!(call[cursor]==IntHarvester.PREDICATE||(!predicateOnly&&call[cursor]==IntHarvester.NUMBER))){
			if(bi.p.trace)System.out.println("Argument is not a correct."); 
			return false;}// must start with a predicate
		return true;
	}
	/**
	 * Assert an array to the rules.
	 * @param array Array to be asserted.
	 * @return Whether assert was successful.
	 */
	public boolean assertArray(int[] array){
		PrimIntKeyHashMap<PrimIntArrayList> byName = bi.p.rulebase.getObj(array[1]); // get entries by ID
		PrimIntArrayList rules = null;
		if(byName == null){ // no entries yet for id
			byName = new PrimIntKeyHashMap<PrimIntArrayList>(PrimIntKeyHashMap.OBJ);
			bi.p.rulebase.putObj(array[1], byName);  // add for this id an entry
		} else {
			rules = byName.getObj(array[2]); // get the arity's entries
			if(rules != null){ // if those exist
				for(int i = 0; i < rules.pointer; i++){
					boolean same = true;
					int[] entry = rules.arraydata[i];
					if(entry.length!=array.length) same = false;
					else for(int j = 0; j < entry.length; j++)
						if(entry[j]!=array[j]){
							same = false;
							j = entry.length;
						} 
					if(same)return false; // already exists
				}
				rules.addAr(array);
				return true;
			}
		}
		rules = new PrimIntArrayList(false); // add new rules for this id/arity combo
		rules.addAr(array);
		byName.putObj(array[2], rules);
		return true;
	}

	/**
	 * Retract an array from the rules.
	 * @param array Array to be retracted.
	 * @return Whether retract was successful.
	 */
	public boolean retractArray(int[] array){
		//{on(maillist,Person)}Unsubscribe(Person){not on(maillist,Person), not address(Person,_)} // is annoying this isn't possible
		PrimIntKeyHashMap<PrimIntArrayList> byName = bi.p.rulebase.getObj(array[1]); // get entries by ID
		if(byName != null){ 
			PrimIntArrayList rules = byName.getObj(array[2]); // get the arity's entries
			if(rules != null){ // if those exist
				for(int i = 0; i < rules.pointer; i++){ // go through the entries to search for a match
					int[] entry = rules.arraydata[i];
					if(entry.length==array.length){ // at least length must be the same
						boolean same = true;
						for(int j = 0; j < entry.length; j++) // go through ints
							if(entry[j]!=array[j]){ 
								same = false;
								j = entry.length;
							} 
						if(same){ // they are the same
							rules.removeArAt(i); // remove the array 
							return true;
						}
					}
				} 
			}
		}
		return false;
	}

	/**
	 * Deep copy an element (include integers that are part of variable values).
	 * @param call Array which needs to be deep copied.
	 * @param start Start of the term in the array.
	 * @return Deep copy of the term starting at "start".
	 */
	public int[] deepcopy(int[] call, int start, int end){
		array.clear(); // clear out memory
		vars.clear();
		splits.clear();
		varIndices.clear();
		int cursor = start;

		if(addZeroSplit)splits.addInt(0); // makes the term a query, for 2OPL
		while(cursor!=end){
			cursor = deepcopyScan(call,cursor); // obtain the integers in the array
			splits.addInt(array.pointer); // split overhead
		}

		// Put variable overhead in the array
		for(int var = vars.size()-1; var >= 0; var--){
			PrimIntArrayList indices = varIndices.getObj(var);
			for(int i = indices.pointer-1; i >= 0; i--) array.addInt(indices.intdata[i]);
			array.addInt(indices.pointer);
		}
		array.addInt(vars.size());

		// Put split overhead in the array
		for(int split = splits.pointer-1; split >= 0; split--) 
			array.addInt(splits.intdata[split]);
		array.addInt(splits.pointer);
		return bi.p.listToPrim(array);
	}
	/**
	 * Scan over an element and copy all the integers (combination of IntHarvester.scanElement and the parser).
	 * @param rule Array to scanned.
	 * @param start Start of the term.
	 * @return End integer of the term.
	 */
	public int deepcopyScan(int[] rule, int start){
		int type = rule[start];
		if(type==IntHarvester.PREDICATE){
			int x = start+3; // start of args
			int nrargs = rule[start+2];
			array.addInt(rule[start]);array.addInt(rule[start+1]);array.addInt(rule[start+2]);
			for(int i = 0; i < nrargs;i++) x = deepcopyScan(rule,x); // scan over arguments
			return x;
		} else if(type==IntHarvester.VARIABLE){
			int id = rule[start+1];
			int[] value = bi.p.substitution.value(id);
			if(value==null){ // add variable to the rule
				id = vars.getInt(id);
				if(id < 0){
					id = vars.size();
					vars.putInt(rule[start+1], id);
				}
				array.addInt(rule[start]); // add the type integer
				PrimIntArrayList spots = varIndices.getObj(id);
				if(spots == null){
					spots = new PrimIntArrayList(true);
					varIndices.putObj(id, spots);
				}
				spots.addInt(array.pointer); // add variable overhead
				array.addInt(id);  // add the id to the array
			} else deepcopyScan(value, 0); // else copy the inner value
			return start+2;
		} else if(type==IntHarvester.LIST){
			int x = start+2; // start of elems
			int nrElems = rule[start+1];
			array.addInt(rule[start]);array.addInt(rule[start+1]);
			for(int i = 0; i < nrElems;i++) x = deepcopyScan(rule,x); // scan over elems
			if(nrElems>0)x = deepcopyScan(rule,x);// also skip tail
			return x; 
		} else if(type==IntHarvester.NUMBER){
			array.addInt(rule[start]);array.addInt(rule[start+1]);array.addInt(rule[start+2]);
			return start+3;
		} else if(type==IntHarvester.PARENTHESIZED){
			int x = start+2; // start of args
			int nrElems = rule[start+1];
			array.addInt(rule[start]);array.addInt(rule[start+1]);
			for(int i = 0; i < nrElems;i++) x = deepcopyScan(rule,x); // scan over arguments
			return x;
		} else if(type==IntHarvester.CUT){ 
			array.addInt(rule[start]);array.addInt(0);
			return start+2;
		} else {
			return -1;
		}
	}
}
