package networkConnection;

import java.net.*;
import java.sql.ClientInfoStatus;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.io.*;
public class Server extends Thread{

	// Attributen
	private Socket clientsocket;
	private ServerSocket serversocket;
	private ObjectOutputStream clientOOS;
	private ObjectInputStream clientOIS;
	private int serverPort = 0;
	private ClientMessageHandler cmh;
	private ArrayList<NetworkListener> listeners = new ArrayList<NetworkListener>();
	Calendar cal = Calendar.getInstance();
	SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
	
	public Server(int serverPort, ClientMessageHandler cmh){
		this.serverPort = serverPort;
		this.cmh = cmh;
	}
	
	public void addListener(NetworkListener nl){
		listeners.add(nl);
	}
	
	public void removeListener(NetworkListener nl){
		listeners.remove(nl);
	}
	
	/**
	 * Connect to a server.
	 * @param host The server's host adress.
	 * @param port The port to connect to.
	 */
	public void connectToServer(String host, int port){
		try { // make the required objects
			clientsocket = new Socket(host, port); // make the socket
			clientOOS = new ObjectOutputStream(clientsocket.getOutputStream()); // get the output stream
			clientOIS = new ObjectInputStream(clientsocket.getInputStream()); // get the input stream
		}
		catch(UnknownHostException ue) {
            outputEvent("Exception:\r\n"+ue.getMessage());
		}
		catch(IOException ie) {
            outputEvent("Exception:\r\n"+ie.getMessage());
		}	
	}

	/**
	 * Disconnect from the server.
	 */
	public void disconnectFromServer() {
		try {
			// Close the socket and streams.
			clientsocket.close();
			clientOOS.close();
		}
		catch(IOException ie) {
            outputEvent("Exception:\r\n"+ie.getMessage());
		}
	}

	/**
	 * Connect to a server and then send it a message.
	 * @param msg The message.
	 * @param host Server host.
	 * @param port Server port.
	 * @param waitForAnswer Block everything for answer.
	 * @param disconnectAfterwards Whether to disconnect after the message is send.
	 * @return The answer object.
	 */
	public Object sendMessage(Object msg, String host, int port, boolean waitForAnswer, boolean disconnectAfterwards) {
		connectToServer(host, port);
		return sendMessage(msg, waitForAnswer, disconnectAfterwards);
	}
	
	/**
	 * Send the server a message.
	 * @param msg The message.
	 * @param waitForAnswer Block everything for answer.
	 * @param disconnectAfterwards Whether to disconnect after the message is send.
	 * @return The answer object.
	 */
	public Object sendMessage(Object msg, boolean waitForAnswer, boolean disconnectAfterwards) {
		Object answer = null;
		try { 
			clientOOS.writeObject(msg);
			clientOOS.flush();
			outputMessageEvent(InetAddress.getLocalHost().getHostAddress(),clientsocket.getInetAddress().getHostAddress(), msg, false);
			if(waitForAnswer){
				answer = clientOIS.readObject();
				outputMessageEvent(clientsocket.getInetAddress().getHostAddress(), InetAddress.getLocalHost().getHostAddress(), answer, true);
			}
		}
		catch(Exception ie) {
            outputEvent("Exception:\r\n"+ie.getMessage());
		}
		if(disconnectAfterwards) disconnectFromServer();
		return answer;
	}
	/*
	 * Run method for server side of this objects.
	 */
	public void run(){
        try {
            serversocket = new ServerSocket(serverPort); // create the server socket
            outputEvent("Server established connection at port " + serverPort);
            while (!serversocket.isClosed()) {
                // Wait for client call
                try {Socket connectionsocket = serversocket.accept(); // accept the call
                ObjectOutputStream oos = new ObjectOutputStream(connectionsocket.getOutputStream());
                oos.flush();
                ObjectInputStream ois = new ObjectInputStream(connectionsocket.getInputStream());
                cmh.handle(connectionsocket,ois,oos, ois.readObject()); // let the program handle the message
    			oos.close();
    			ois.close();
                connectionsocket.close();
                }catch(Exception e){}
            }
        }
        catch(BindException ex) {
            outputEvent("Failed to connect the server to port "+serverPort+".");
        }
        catch(Exception ex){
            outputEvent("Exception:\r\n"+ex.getMessage());
        }
    }
	
	/**
	 * Disconnect the server socket.
	 */
	public void disconnectServer(){
        try {
            serversocket.close();
            outputEvent("Server connection at port " + serverPort + " is closed.");
        } catch (IOException ex) {
        	outputEvent("Connection at port " + serverPort + " could not be closed:\r\n" + ex.getMessage());
        } catch (Exception ex) {
            outputEvent("Exception:\r\n"+ex.getMessage());
        }
        
    }
	
	public void outputEvent(String s){
		String date = sdf.format(cal.getTime());
        for(NetworkListener nl : listeners) 
        	nl.notifyOverheadEvent(date,s);
	}
	
	public void outputMessageEvent(String from, String to, Object message, boolean received){
		String date = sdf.format(cal.getTime());
		for(NetworkListener nl : listeners)
			nl.notifyMessageTransfer(date, from, to, message, received);
	}
}
